#!/usr/bin/env python3

# stdlib modules
import json
import logging
import copy

# third party imports
import numpy as np
from openquake.hazardlib.geo.point import Point

# local imports
from .point_rupture import PointRupture
from .quad_rupture import QuadRupture
from .edge_rupture import EdgeRupture
from .origin import Origin
from .utils import is_quad
from .constants import TIMEFMT, ALT_TIMEFMT, DEPTH_TOL
from esi_utils_time.ancient_time import HistoricTime


def get_rupture(origin, file=None, mesh_dx=0.5, new_format=True):
    """
    This is a module-level function to read in a rupture file. This allows for
    the ShakeMap 3 text file specification or the ShakeMap 4 JSON rupture
    format. The ShakeMap 3 (".txt" extension) only supports QuadRupture style
    rupture representation and so this method will always return a QuadRupture
    instance. The ShakeMap 4 JSON format supports QuadRupture and EdgeRupture
    represenations and so this method detects the rupture class and returns the
    appropriate Rupture subclass instance.

    If file is None (default) then it returns a PointRupture.

    Args:
        origin (Origin):
            A ShakeMap origin instance; required because hypocentral/epicentral
            distances are computed from the Ruptureclass.

        file (srt):
            Path to rupture file (optional).

        mesh_dx (float):
            Target spacing (in km) for rupture discretization; default is
            0.5 km and it is only used if the rupture file is an EdgeRupture.

        new_format (bool):
            Indicates whether text rupture format is
            "old" (lat, lon, depth) or "new" (lon, lat, depth) style.

    Returns:
        Rupture subclass instance.

    """
    if not isinstance(origin, Origin):
        raise TypeError("An Origin is requred.")

    if file is not None:
        try:
            # -----------------------------------------------------------------
            # First, try to read as a json file
            # -----------------------------------------------------------------
            if isinstance(file, str):
                with open(file, encoding="latin-1") as f:
                    d = json.load(f)
            else:
                d = json.loads(str(file))

            rupt = rupture_from_dict_and_origin(d, origin, mesh_dx=mesh_dx)

        except Exception as e:
            if not isinstance(e, json.JSONDecodeError) and not isinstance(
                e, UnicodeDecodeError
            ):
                logging.warning(f"Unknown exception reading fault file: {str(e)}")
            # -----------------------------------------------------------------
            # Reading as json failed, so hopefully it is a ShakeMap 3 text file
            # -----------------------------------------------------------------
            try:
                d = text_to_json(file, new_format=new_format)
                rupt = rupture_from_dict_and_origin(d, origin, mesh_dx=mesh_dx)
            except ValueError as e:
                logging.error(e)
                raise ValueError(
                    f"Error reading {file}. This could have been "
                    "because the latitude and longitdue are "
                    "reversed. Try changing the 'new_format' "
                    "option."
                )
            except Exception as e:
                logging.error(e)
                raise IOError("Unknown rupture file format.")
    else:
        rupt = PointRupture(origin)

    return rupt


def rupture_from_dict_and_origin(rupdict, origin, mesh_dx=0.5):
    """
    Method returns either a QuadRupture or EdgeRupture object based on a
    GeoJSON dictionary and an origin. Note that this is very similar to
    :func:`rupture_from_dict`, except that method is for
    constructing the rupture objects from a dict that already contains the
    origin info in the `metadata` field (e.g., from a dict from a Shakemap
    container), while this method is for construction of the rupture objects
    from a GeoJSON dict that does not yet include that information (e.g., from
    a dict that is read in to initially create the shakemap container, along
    with an Origin that is derived from `event.xml`).

    .. seealso:: :func:`rupture_from_dict`

    Args:
        rupdictd (dict):
            Rupture GeoJSON dictionary.
        origin (Origin):
            A ShakeMap origin object.
        mesh_dx (float):
            Target spacing (in km) for rupture discretization;
            default is 0.5 km and it is only used if the rupture file is an
            EdgeRupture.

    Returns:
        a Rupture subclass.

    """
    validate_json(rupdict)

    # Is this a QuadRupture or an EdgeRupture?
    valid_quads = is_quadrupture_class(rupdict)

    if valid_quads is True:
        rupt = QuadRupture(rupdict, origin)
    else:
        if rupdict["features"][0]["geometry"]["type"] == "Point":
            rupt = PointRupture(origin)
        else:
            rupt = EdgeRupture(rupdict, origin, mesh_dx=mesh_dx)

    return rupt


def rupture_from_dict(d):
    """
    Method returns either a Rupture subclass (QuadRupture, EdgeRupture, or
    PointRupture) object based on a GeoJSON dictionary.

    .. seealso::
        :func:`rupture_from_dict_and_origin`

    Args:
        d (dict):
            Rupture GeoJSON dictionary, which must contain origin
            information in the 'metadata' field.

    Returns:
        a Rupture subclass.

    """
    validate_json(d)

    # We don't want to mess with the input just in case it gets used again
    d = copy.deepcopy(d)

    try:
        d["metadata"]["time"] = HistoricTime.strptime(d["metadata"]["time"], TIMEFMT)
    except ValueError:
        d["metadata"]["time"] = HistoricTime.strptime(
            d["metadata"]["time"], ALT_TIMEFMT
        )

    origin = Origin(d["metadata"])

    # What type of rupture is this?
    geo_type = d["features"][0]["geometry"]["type"]
    if geo_type == "MultiPolygon":
        # EdgeRupture will have 'mesh_dx' in metadata
        if "mesh_dx" in d["metadata"]:
            mesh_dx = d["metadata"]["mesh_dx"]
            rupt = EdgeRupture(d, origin, mesh_dx=mesh_dx)
        else:
            rupt = QuadRupture(d, origin)
    elif geo_type == "Point":
        rupt = PointRupture(origin)

    return rupt


def _rotate_polygon(p):
    """
    This is a method to rotate a polygon, which is used to try to
    fix incorrectly specified polygons.

    Args:
        p (list):
            A list of lon/lat/depth lists.

    Returns:
        list: Same as before but the points have been incremented
        by one in position.

    """
    # Convert to numpy array for better indexing support
    parray = np.array(p)

    # Ensure the polygon is closed
    if not np.array_equal(parray[-1, :], parray[0, :]):
        raise Exception("Rupture file has unclosed segments.")

    # Drop last point
    parray = parray[0:-1, :]

    # Rotate by putting first at end
    parray = np.append(parray[1:, :], np.reshape(parray[0, :], (1, -1)), axis=0)

    # Put new first point onto the end so that it is closed.
    parray = np.append(parray, np.reshape(parray[0, :], (1, -1)), axis=0)

    # Turn numpy array back into a list
    polygon = parray.tolist()
    return polygon


def _check_polygon(p):
    """
    Check if the verticies are specified top first.

    Args:
        p (list):
            A list of five lon/lat/depth lists.

    Raises:
        ValueError: incorrectly specified polygon.

    """
    n_points = len(p)
    if n_points % 2 == 0:
        raise ValueError("Number of points in polyon must be odd.")

    if p[0] != p[-1]:
        raise ValueError("First and last points in polygon must be " "identical.")

    n_pairs = int((n_points - 1) / 2)
    for j in range(n_pairs):
        # -------------------------------------------------------------
        # Points are paired and in each pair the top is first, as in:
        #
        #      _.-P1-._
        #   P0'        'P2---P3
        #   |                  \
        #   P7---P6----P5-------P4
        #
        # Pairs: P0-P7, P1-P6, P2-P5, P3-P4
        # -------------------------------------------------------------
        top_depth = p[j][2]
        bot_depth = p[-(j + 2)][2]
        if top_depth >= bot_depth:
            raise ValueError("Top points must be ordered before bottom points.")


def text_to_json(file, new_format=True):
    """
    Read in old or new ShakeMap 3 textfile rupture format and convert to
    GeoJSON.

    This will handle ShakeMap3.5-style fault text files, which can have the
    following format:
     - # at the top indicates a reference.
     - Lines beginning with a > indicate the end of one segment and the
       beginning of another.
     - Coordinates are specified in lat,lon,depth order.
     - Coordinates can be separated by commas or spaces.
     - Vertices can be specified in top-edge or bottom-edge first order.

    Args:
        file (str):
            Path to rupture file OR file-like object in GMT
            psxy format, where:

                * Rupture vertices are space/comma separated lat, lon, depth
                  triplets on a single line.
                * Rupture groups are separated by lines containing ">"
                * Rupture groups must be closed.
                * Verticies within a rupture group must start along the top
                  edge and move in the strike direction then move to the bottom
                  edge and move back in the opposite direction.

        new_format (bool):
            Indicates whether text rupture format is
            "old" (lat, lon, depth) or "new" (lon, lat, depth) style.

    Returns:
        dict: GeoJSON rupture dictionary.

    """
    isfile = False
    if hasattr(file, "read"):
        f = file
    else:
        f = open(file, "rt", encoding="latin-1")
        isfile = True

    reference = ""
    polygons = []
    polygon = []
    for line in f.readlines():
        if not len(line.strip()):
            continue

        if line.strip().startswith("#"):
            # Get reference string
            reference += line.strip().replace("#", "")
            continue

        if line.strip().startswith(">"):
            if not len(polygon):
                continue
            polygons.append(polygon)
            polygon = []
            continue

        # first try to split on whitespace
        parts = line.split()
        if len(parts) == 1:
            if new_format:
                raise Exception(f"Rupture file {file} has unspecified delimiters.")
            parts = line.split(",")
            if len(parts) == 1:
                raise Exception(f"Rupture file {file} has unspecified delimiters.")

        if len(parts) != 3:
            msg = f"Rupture file {file} is not in lat, lon, depth format."
            if new_format:
                f"Rupture file {file} is not in lon, lat, depth format."
            raise Exception(msg)

        parts = [float(p) for p in parts]
        if not new_format:
            old_parts = parts.copy()
            parts[0] = old_parts[1]
            parts[1] = old_parts[0]
        polygon.append(parts)

    if len(polygon):
        polygons.append(polygon)

    if isfile:
        f.close()

    # Try to fix polygons
    original_polygons = polygons.copy()
    fixed = []
    n_polygons = len(polygons)
    for i in range(n_polygons):
        n_verts = len(polygons[i])
        success = False
        for j in range(n_verts - 1):
            try:
                _check_polygon(polygons[i])
                success = True
                break
            except ValueError:
                polygons[i] = _rotate_polygon(polygons[i])
        if success:
            fixed.append(True)
        else:
            fixed.append(False)

    if not all(fixed):
        polygons = original_polygons

    json_dict = {
        "type": "FeatureCollection",
        "metadata": {"reference": reference},
        "features": [
            {
                "type": "Feature",
                "properties": {"rupture type": "rupture extent"},
                "geometry": {"type": "MultiPolygon", "coordinates": [polygons]},
            }
        ],
    }
    validate_json(json_dict)

    return json_dict


def validate_json(d):
    """
    Check that the JSON format is acceptable. This is only for requirements
    that are common to both QuadRupture and EdgeRupture.

    Args:
        d (dict): Rupture JSON dictionary.
    """
    if d["type"] != "FeatureCollection":
        raise Exception('JSON file is not a "FeatureColleciton".')

    if len(d["features"]) != 1:
        raise Exception("JSON file should contain excactly one feature.")

    if "reference" not in d["metadata"].keys():
        raise Exception("Json metadata field should contain " '"reference" key.')

    f = d["features"][0]

    if f["type"] != "Feature":
        raise Exception('Feature type should be "Feature".')

    geom = f["geometry"]

    if geom["type"] != "MultiPolygon" and geom["type"] != "Point":
        raise Exception('Geometry type should be "MultiPolygon" ' 'or "Point".')

    if "coordinates" not in geom.keys():
        raise Exception('Geometry dictionary should contain "coordinates" ' "key.")

    polygons = geom["coordinates"][0]

    if geom["type"] == "MultiPolygon":
        n_polygons = len(polygons)
        for i in range(n_polygons):
            _check_polygon(polygons[i])


def is_quadrupture_class(d):
    """
    Check if JSON file fulfills QuadRupture class criteria:

        - Are top and bottom edges horizontal?
        - Are the four points in each quad coplanar?

    Args:
        d (dict): Rupture JSON dictionary.

    Returns:
        bool: Can the rupture be represented in the QuadRupture class?
    """
    f = d["features"][0]
    geom = f["geometry"]
    if geom["type"] == "Point":
        return False
    polygons = geom["coordinates"][0]
    try:
        len(polygons)
    except Exception:
        return False
    n_polygons = len(polygons)
    for i in range(n_polygons):
        p = polygons[i]
        n_points = len(p)
        n_pairs = int((n_points - 1) / 2)
        n_quads = n_pairs - 1

        for k in range(n_quads):
            # Four points of each quad should be co-planar within a tolerance
            quad = [
                Point(p[k][0], p[k][1], p[k][2]),
                Point(p[k + 1][0], p[k + 1][1], p[k + 1][2]),
                Point(p[-(k + 3)][0], p[-(k + 3)][1], p[-(k + 3)][2]),
                Point(p[-(k + 2)][0], p[-(k + 2)][1], p[-(k + 2)][2]),
            ]
            test = is_quad(quad)
            if test[0] is False:
                return False

            # Within each quad, top and bottom edges must be horizontal
            tops = np.array([quad[0].depth, quad[1].depth])
            if not np.isclose(tops[0], tops, rtol=0, atol=DEPTH_TOL).all():
                return False
            bots = np.array([quad[2].depth, quad[3].depth])
            if not np.isclose(bots[0], bots, rtol=0, atol=DEPTH_TOL).all():
                return False

    return True



#!/usr/bin/env python3

# stdlib modules
import itertools as it
import copy

# third party imports
import numpy as np
from openquake.hazardlib.geo.utils import OrthographicProjection
from openquake.hazardlib.geo import geodetic

# local imports
from esi_utils_vectors.vector import Vector
from .utils import reverse_quad, get_quad_length


def _computeGC2(rupture, lon, lat, depth):
    """
    Method for computing GC2 from a ShakeMap Rupture instance.

    Args:
        rupture (Rupture): ShakeMap rupture object.
        lon (array): Numpy array of site longitudes.
        lat (array): Numpy array of site latitudes.
        depth (array): Numpy array of site depths.

    Returns:
        dict: Dictionary of GC2 distances. Keys include "T", "U", "rx"
            "ry", "ry0".
    """

    quadlist = rupture.getQuadrilaterals()
    quadgc2 = copy.deepcopy(quadlist)

    oldshape = lon.shape

    if len(oldshape) == 2:
        newshape = (oldshape[0] * oldshape[1], 1)
    else:
        newshape = (oldshape[0], 1)

    # -------------------------------------------------------------------------
    # Define a projection that spans sites and rupture
    # -------------------------------------------------------------------------

    all_lat = np.append(lat, rupture.lats)
    all_lon = np.append(lon, rupture.lons)

    west = np.nanmin(all_lon)
    east = np.nanmax(all_lon)
    south = np.nanmin(all_lat)
    north = np.nanmax(all_lat)
    proj = OrthographicProjection(west, east, north, south)

    totweight = np.zeros(newshape, dtype=lon.dtype)
    GC2T = np.zeros(newshape, dtype=lon.dtype)
    GC2U = np.zeros(newshape, dtype=lon.dtype)

    # -------------------------------------------------------------------------
    # First sort out strike discordance and nominal strike prior to
    # starting the loop if there is more than one group/trace.
    # -------------------------------------------------------------------------
    group_ind = rupture._getGroupIndex()

    # Need group_ind as numpy array for sensible indexing...
    group_ind_np = np.array(group_ind)
    uind = np.unique(group_ind_np)
    n_groups = len(uind)

    if n_groups > 1:
        # ---------------------------------------------------------------------
        # The first thing we need to worry about is finding the coordinate
        # shift. U's origin is "selected from the two endpoints most
        # distant from each other."
        # ---------------------------------------------------------------------

        # Need to get index of first and last quad
        # for each segment
        iq0 = np.zeros(n_groups, dtype="int16")
        iq1 = np.zeros(n_groups, dtype="int16")
        for k in uind:
            ii = [i for i, j in enumerate(group_ind) if j == uind[k]]
            iq0[k] = int(np.min(ii))
            iq1[k] = int(np.max(ii))

        # ---------------------------------------------------------------------
        # This is an iterator for each possible combination of traces
        # including trace orientations (i.e., flipped).
        # ---------------------------------------------------------------------

        it_seg = it.product(it.combinations(uind, 2), it.product([0, 1], [0, 1]))

        # Placeholder for the trace pair/orientation that gives the
        # largest distance.
        dist_save = 0

        for k in it_seg:
            s0ind = k[0][0]
            s1ind = k[0][1]
            p0ind = k[1][0]
            p1ind = k[1][1]
            if p0ind == 0:
                P0 = quadlist[iq0[s0ind]][0]
            else:
                P0 = quadlist[iq1[s0ind]][1]
            if p1ind == 0:
                P1 = quadlist[iq1[s1ind]][0]
            else:
                P1 = quadlist[iq0[s1ind]][1]

            dist = geodetic.distance(
                P0.longitude, P0.latitude, 0.0, P1.longitude, P1.latitude, 0.0
            )
            if dist > dist_save:
                dist_save = dist
                A0 = P0
                A1 = P1

        # ---------------------------------------------------------------------
        # A0 and A1 are the furthest two segment endpoints, but we still
        # need to sort out which one is the "origin".
        # ---------------------------------------------------------------------

        # This goofy while-loop is to adjust the side of the rupture where the
        # origin is located
        dummy = -1
        while dummy < 0:
            A0.depth = 0
            A1.depth = 0
            p_origin = Vector.fromPoint(A0)
            a0 = Vector.fromPoint(A0)
            a1 = Vector.fromPoint(A1)
            ahat = (a1 - a0).norm()

            # Loop over traces
            e_j = np.zeros(n_groups)
            b_prime = [None] * n_groups
            for j in range(n_groups):
                P0 = quadlist[iq0[j]][0]
                P1 = quadlist[iq1[j]][1]
                P0.depth = 0
                P1.depth = 0
                p0 = Vector.fromPoint(P0)
                p1 = Vector.fromPoint(P1)
                b_prime[j] = p1 - p0
                e_j[j] = ahat.dot(b_prime[j])
            E = np.sum(e_j)

            # List of discordancy
            dc = [np.sign(a) * np.sign(E) for a in e_j]
            b = Vector(0, 0, 0)
            for j in range(n_groups):
                b.x = b.x + b_prime[j].x * dc[j]
                b.y = b.y + b_prime[j].y * dc[j]
                b.z = b.z + b_prime[j].z * dc[j]
            bhat = b.norm()
            dummy = bhat.dot(ahat)
            if dummy < 0:
                tmpA0 = copy.deepcopy(A0)
                A0 = copy.deepcopy(A1)
                A1 = tmpA0

        # ---------------------------------------------------------------------
        # To fix discordancy, need to flip quads and rearrange
        # the order of quadgc2
        # ---------------------------------------------------------------------

        # 1) flip quads
        for i in range(len(quadgc2)):
            if dc[group_ind[i]] < 0:
                quadgc2[i] = reverse_quad(quadgc2[i])

        # 2) rearrange quadlist order
        qind = np.arange(len(quadgc2))
        for i in range(n_groups):
            qsel = qind[group_ind_np == uind[i]]
            if dc[i] < 0:
                qrev = qsel[::-1]
                qind[group_ind_np == uind[i]] = qrev

        quadgc2old = copy.deepcopy(quadgc2)
        for i in range(len(qind)):
            quadgc2[i] = quadgc2old[qind[i]]

        # End of if-statement for adjusting group discordancy

    s_i = 0.0
    l_i = np.zeros(len(quadgc2))

    for i in range(len(quadgc2)):
        G0, G1, G2, G3 = quadgc2[i]

        # Compute u_i and t_i for this quad
        t_i = __calc_t_i(G0, G1, lat, lon, proj)
        u_i = __calc_u_i(G0, G1, lat, lon, proj)

        # Quad length (top edge)
        l_i[i] = get_quad_length(quadgc2[i])

        # ---------------------------------------------------------------------
        # Weight of segment, three cases
        # ---------------------------------------------------------------------

        # Case 3: t_i == 0 and 0 <= u_i <= l_i
        w_i = np.zeros_like(t_i)

        # To avoid division by zero in totweight later on:
        ix = (t_i == 0) & (0 <= u_i) & (u_i <= l_i[i])
        totweight[ix] = 1.0

        # Case 1:
        ix = t_i != 0
        w_i[ix] = (1.0 / t_i[ix]) * (
            np.arctan((l_i[i] - u_i[ix]) / t_i[ix]) - np.arctan(-u_i[ix] / t_i[ix])
        )

        # Case 2:
        ix = (t_i == 0) & ((u_i < 0) | (u_i > l_i[i]))
        w_i[ix] = 1 / (u_i[ix] - l_i[i]) - 1 / u_i[ix]

        totweight = totweight + w_i
        GC2T = GC2T + w_i * t_i

        if n_groups == 1:
            GC2U = GC2U + w_i * (u_i + s_i)
        else:
            if i == 0:
                qind = np.array(range(len(quadgc2)))
                l_kj = 0
                s_ij_1 = 0
            else:
                l_kj = l_i[(group_ind_np == group_ind_np[i]) & (qind < i)]
                s_ij_1 = np.sum(l_kj)

            # First endpoint in the current 'group' (or 'trace' in GC2 terms)
            p1 = Vector.fromPoint(quadgc2[iq0[group_ind[i]]][0])
            s_ij_2 = (p1 - p_origin).dot(np.sign(E) * ahat) / 1000.0

            # Above is GC2N, for GC2T use:
            # s_ij_2 = (p1 - p_origin).dot(bhat) / 1000.0

            s_ij = s_ij_1 + s_ij_2
            GC2U = GC2U + w_i * (u_i + s_ij)

        s_i = s_i + l_i[i]

    GC2T = GC2T / totweight
    GC2U = GC2U / totweight

    # Dictionary for holding the distances
    distdict = dict()

    distdict["T"] = copy.deepcopy(GC2T).reshape(oldshape)
    distdict["U"] = copy.deepcopy(GC2U).reshape(oldshape)

    # Take care of Rx
    Rx = copy.deepcopy(GC2T)  # preserve sign (no absolute value)
    Rx = Rx.reshape(oldshape)
    distdict["rx"] = Rx

    # Ry
    Ry = GC2U - s_i / 2.0
    Ry = Ry.reshape(oldshape)
    distdict["ry"] = Ry

    # Ry0
    Ry0 = np.zeros_like(GC2U)
    ix = GC2U < 0
    Ry0[ix] = np.abs(GC2U[ix])
    if n_groups > 1:
        s_i = s_ij + l_i[-1]
    ix = GC2U > s_i
    Ry0[ix] = GC2U[ix] - s_i
    Ry0 = Ry0.reshape(oldshape)
    distdict["ry0"] = Ry0

    return distdict


def __calc_u_i(P0, P1, lat, lon, proj):
    """
    Calculate u_i distance. See Spudich and Chiou OFR 2015-1028. This is the
    distance along strike from the first vertex (P0) of the i-th segment.

    Args:
        P0 (point): OQ Point object, representing the first top-edge vertex of
            a rupture quadrilateral.
        P1 (point): OQ Point object, representing the second top-edge vertex of
            a rupture quadrilateral.
        lat (array): A numpy array of latitudes.
        lon (array): A numpy array of longitudes.
        proj (object): An orthographic projection from
            openquake.hazardlib.geo.utils.OrthographicProjection.

    Returns:
        array: Array of size N of distances (in km) from input points to
            rupture surface.

    """

    # projected coordinates are in km
    p0x, p0y = proj(P0.x, P0.y)
    p1x, p1y = proj(P1.x, P1.y)

    # Unit vector pointing along strike
    u_i_hat = Vector(p1x - p0x, p1y - p0y, 0).norm()

    # Convert sites to Cartesian
    sx, sy = proj(lon, lat)
    sx1d = np.reshape(sx, (-1,))
    sy1d = np.reshape(sy, (-1,))

    # Vectors from P0 to sites
    r = np.zeros([len(sx1d), 2])
    r[:, 0] = sx1d - p0x
    r[:, 1] = sy1d - p0y

    # Dot product gives u_i
    u_i = np.sum(u_i_hat.getArray()[0:2] * r, axis=1)
    shp = u_i.shape
    if len(shp) == 1:
        u_i.shape = (shp[0], 1)
    u_i = np.fliplr(u_i)

    return u_i


def __calc_t_i(P0, P1, lat, lon, proj):
    """
    Calculate t_i distance. See Spudich and Chiou OFR 2015-1028. This is the
    distance measured normal to strike from the i-th segment. Values on the
    hanging-wall are positive and those on the foot-wall are negative.

    Args:
        P0 (point): OQ Point object, representing the first top-edge vertex of
            a rupture quadrilateral.
        P1 (point): OQ Point object, representing the second top-edge vertex of
            a rupture quadrilateral.
        lat (array): A numpy array of latitudes.
        lon (array): A numpy array of longitudes.
        proj (object): An orthographic projection from
            openquake.hazardlib.geo.utils.OrthographicProjection.

    Returns:
        array: Array of size N of distances (in km) from input points to
            rupture surface.
    """

    # projected coordinates are in km
    p0x, p0y = proj(P0.x, P0.y)
    p1x, p1y = proj(P1.x, P1.y)

    # Unit vector pointing normal to strike
    t_i_hat = Vector(p1y - p0y, -(p1x - p0x), 0).norm()

    # Convert sites to Cartesian
    sx, sy = proj(lon, lat)
    sx1d = np.reshape(sx, (-1,))
    sy1d = np.reshape(sy, (-1,))

    # Vectors from P0 to sites
    r = np.zeros([len(sx1d), 2])
    r[:, 0] = sx1d - p0x
    r[:, 1] = sy1d - p0y

    # Dot product gives t_i
    t_i = np.sum(t_i_hat.getArray()[0:2] * r, axis=1)
    shp = t_i.shape
    if len(shp) == 1:
        t_i.shape = (shp[0], 1)
    t_i = np.fliplr(t_i)
    return t_i



# third party imports
import numpy as np
from obspy.imaging.beachball import aux_plane, mt2axes, mt2plane, MomentTensor


def fill_tensor_from_angles(strike, dip, rake, magnitude=6.0, source='unknown',
                            mtype='unknown'):
    """Fill in moment tensor parameters from strike,dip, and rake.

    Args:
        strike (float): Strike from (assumed) first nodal plane (degrees).
        dip (float): Dip from (assumed) first nodal plane (degrees).
        rake (float): Rake from (assumed) first nodal plane (degrees).
        magnitude (float): Magnitude for moment tensor
            (not required if using moment tensor for angular comparisons.)
        source (str): Source (network, catalog) for input parameters.
        mtype (str): Focal mechanism or moment tensor type (Mww,Mwb,Mwc, etc.)
    Returns:
        dict: Fully descriptive moment tensor dictionary, including fields:
            - mrr,mtt,mpp,mrt,mrp,mtp Moment tensor components.
            - T T-axis values:
              - azimuth (degrees)
              - plunge (degrees)
            - N N-axis values:
              - azimuth (degrees)
              - plunge (degrees)
            - P P-axis values:
              - azimuth (degrees)
              - plunge (degrees)
            - NP1 First nodal plane values:
              - strike (degrees)
              - dip (degrees)
              - rake (degrees)
            - NP2 Second nodal plane values:
              - strike (degrees)
              - dip (degrees)
              - rake (degrees)
    """
    mt = plane_to_tensor(strike, dip, rake, mag=magnitude)
    return fill_tensor_from_components(mt[0][0], mt[1][1], mt[2][2],
                                       mt[1][0], mt[0][2], mt[1][2],
                                       source=source, mtype=mtype)


def fill_tensor_from_components(mrr, mtt, mpp, mrt, mrp, mtp, source='unknown',
                                mtype='unknown'):
    """Fill in moment tensor parameters from moment tensor components.

    Args:
        mrr (float): Moment tensor component
        mtt (float): Moment tensor component
        mpp (float): Moment tensor component
        mrt (float): Moment tensor component
        mrp (float): Moment tensor component
        mtp (float): Moment tensor component
        source (str): Source (network, catalog) for input parameters.
        mtype (str): Focal mechanism or moment tensor type (Mww,Mwb,Mwc, etc.)
    Returns:
        dict: Fully descriptive moment tensor dictionary, including fields:
            - mrr,mtt,mpp,mrt,mrp,mtp Moment tensor components.
            - T T-axis values:
              - azimuth (degrees)
              - plunge (degrees)
            - N N-axis values:
              - azimuth (degrees)
              - plunge (degrees)
            - P P-axis values:
              - azimuth (degrees)
              - plunge (degrees)
            - NP1 First nodal plane values:
              - strike (degrees)
              - dip (degrees)
              - rake (degrees)
            - NP2 Second nodal plane values:
              - strike (degrees)
              - dip (degrees)
              - rake (degrees)
    """
    tensor_params = {'mrr': mrr,
                     'mtt': mtt,
                     'mpp': mpp,
                     'mrt': mrt,
                     'mrp': mrp,
                     'mtp': mtp}
    tensor_params['source'] = source
    tensor_params['type'] = mtype
    mt = MomentTensor(mrr, mtt, mpp, mrt, mrp, mtp, 1)
    tnp1 = mt2plane(mt)
    np1 = {'strike': tnp1.strike,
           'dip': tnp1.dip,
           'rake': tnp1.rake}
    tensor_params['NP1'] = np1.copy()
    strike2, dip2, rake2 = aux_plane(np1['strike'], np1['dip'], np1['rake'])
    np2 = {'strike': strike2,
           'dip': dip2,
           'rake': rake2}
    tensor_params['NP2'] = np2.copy()
    T, N, P = mt2axes(mt)
    tensor_params['T'] = {'azimuth': T.strike,
                          'value': T.val,
                          'plunge': T.dip}
    tensor_params['N'] = {'azimuth': N.strike,
                          'value': N.val,
                          'plunge': N.dip}
    tensor_params['P'] = {'azimuth': P.strike,
                          'value': P.val,
                          'plunge': P.dip}

    return tensor_params


def plane_to_tensor(strike, dip, rake, mag=6.0):
    """Convert strike,dip,rake values to moment tensor parameters.

    Args:
        strike (float): Strike from (assumed) first nodal plane (degrees).
        dip (float): Dip from (assumed) first nodal plane (degrees).
        rake (float): Rake from (assumed) first nodal plane (degrees).
        magnitude (float): Magnitude for moment tensor
            (not required if using moment tensor for angular comparisons.)
    Returns:
        nparray: Tensor representation as 3x3 numpy matrix:
            [[mrr, mrt, mrp]
            [mrt, mtt, mtp]
            [mrp, mtp, mpp]]
    """
    # define degree-radian conversions
    d2r = np.pi / 180.0
    # get exponent and moment magnitude
    magpow = mag * 1.5 + 16.1
    mom = np.power(10, magpow)

    # get tensor components
    mrr = mom * np.sin(2 * dip * d2r) * np.sin(rake * d2r)
    mtt = -mom * ((np.sin(dip * d2r) * np.cos(rake * d2r) * np.sin(2 * strike * d2r)) +
                  (np.sin(2 * dip * d2r) * np.sin(rake * d2r) *
                  (np.sin(strike * d2r) * np.sin(strike * d2r))))
    mpp = mom * ((np.sin(dip * d2r) * np.cos(rake * d2r) * np.sin(2 * strike * d2r)) -
                 (np.sin(2 * dip * d2r) * np.sin(rake * d2r) *
                 (np.cos(strike * d2r) * np.cos(strike * d2r))))
    mrt = -mom * ((np.cos(dip * d2r) * np.cos(rake * d2r) * np.cos(strike * d2r)) +
                  (np.cos(2 * dip * d2r) * np.sin(rake * d2r) * np.sin(strike * d2r)))
    mrp = mom * ((np.cos(dip * d2r) * np.cos(rake * d2r) * np.sin(strike * d2r)) -
                 (np.cos(2 * dip * d2r) * np.sin(rake * d2r) * np.cos(strike * d2r)))
    mtp = -mom * ((np.sin(dip * d2r) * np.cos(rake * d2r) * np.cos(2 * strike * d2r)) +
                  (0.5 * np.sin(2 * dip * d2r) * np.sin(rake * d2r) *
                  np.sin(2 * strike * d2r)))

    mt_matrix = np.array([[mrr, mrt, mrp],
                          [mrt, mtt, mtp],
                          [mrp, mtp, mpp]])
    mt_matrix = mt_matrix * 1e-7  # convert from dyne-cm to N-m
    return mt_matrix



#!/usr/bin/env python

# stdlib imports
import warnings

# third party
from lxml import etree
import defusedxml.ElementTree as dET
from openquake.hazardlib.geo import point
from obspy.io.quakeml.core import _is_quakeml
from obspy.core.event import read_events

from esi_utils_rupture.tensor import fill_tensor_from_components
from esi_utils_time.ancient_time import HistoricTime
from esi_utils_rupture import constants


class Origin(object):
    """The purpose of this class is to read/store event origin information,
    which is usually derived from an xml file called "event.xml". There is
    also a mechanism to overwrite information in event.xml with a file in
    the input directory called "source.txt".

    Note that this class is generally contained within a Rupture instance.

    Event values are:

        - id: the event id.
        - netid: the source network code
        - network: a string describing the source network
        - lat: hypocenter latitude (decimal degrees; -90 to 90).
        - lon: hypocenter longitude (decimal degrees; -180 to 180).
        - depth: hypocenter depth (km, positive down).
        - locstring: a free-form descriptive string of location.
        - mag: earthquake magnitude.
        - time: a HistoricTime object containing the origin date/time
        - mech: a string specifying the rupture mechanism; the accepted types
          are RS, SS, NM, and ALL, for reverse slip, strike slip, normal, and
          unspecified ruptures, respectively.
        - reference: A string describing the data source (optional)
        - productcode: The PDL product code for this ShakeMap

    For backward-compatibility, we also check for 'type'. If both 'mech' and
    'type' are missing (or empty strings) then 'mech' is set to ALL.
    """

    def __init__(self, event):
        """Construct an Origin object.

        Args:
            event (dict): Dictionary of values. See list above for required
            keys.

        Returns:
            Origin object.

        Raises:
            KeyError: When one or more required values are missing from the
                      event object.
            TypeError: If the event ID is not a string.
            ValueError: If the latitude or longitude are outside of the
                        allowed ranges, or if the 'mech' value is not
                        an allowed type.
        """

        # ---------------------------------------------------------------------
        # Check for missing keys
        # ---------------------------------------------------------------------
        missing = []
        ekeys = list(event.keys())
        for req in constants.ORIGIN_REQUIRED_KEYS:
            if req not in ekeys:
                missing.append(req)

        if len(missing):
            required_keys = ", ".join(missing)
            raise KeyError(
                "Input event dictionary is missing the following "
                f'required keys: "{required_keys}"'
            )

        # ---------------------------------------------------------------------
        # Check some types, ranges, and defaults
        # ---------------------------------------------------------------------
        if not type(event["id"]) is str and event["id"] is not None:
            raise TypeError("event id  must be a string.")

        if float(event["lat"]) > 90 or float(event["lat"]) < -90:
            raise ValueError("lat must be between -90 and 90 degrees.")

        if float(event["lon"]) > 180 or float(event["lon"]) < -180:
            raise ValueError("lon must be between -180 and 180 degrees.")

        if "type" in ekeys and "mech" not in ekeys:
            event["mech"] = event["type"]
            del event["type"]
        if "mech" in event.keys():
            if event["mech"] == "":
                event["mech"] = constants.DEFAULT_MECH
            if not event["mech"] in constants.RAKEDICT.keys():
                raise ValueError("mech must be SS, NM, RS, or ALL.")
        else:
            event["mech"] = constants.DEFAULT_MECH

        # ---------------------------------------------------------------------
        # Add keys as class attributes
        # ---------------------------------------------------------------------
        for k, v in event.items():
            if k == "rake":
                setattr(self, k, float(v))
            else:
                setattr(self, k, v)

        # What about rake?
        if not hasattr(self, "rake"):
            self.rake = constants.RAKEDICT[self.mech]

        if self.rake is None:
            self.rake = 0.0

        if "reviewed" not in event:
            self.reviewed = "unknown"

        return

    @classmethod
    def fromFile(cls, eventxmlfile, sourcefile=None, momentfile=None):
        """Class method to create a Origin object by specifying an event.xml
        file, a rupture file, a source.txt file, and a moment.xml file.

        Args:
            eventxmlfile (str): Event xml file (see read_event_file()).
            sourcefile (str): source.txt file (see read_source()).
            momentfile (str): moment.xml file (see read_moment_quakeml()).

        Returns:
            Origin object.
        """
        event = read_event_file(eventxmlfile)
        params = None
        if sourcefile is not None:
            params = read_source(sourcefile)
            event.update(params)
        if momentfile is not None:
            params = read_moment_quakeml(momentfile)
            event.update(params)
        return cls(event)

    def getHypo(self):
        """
        Returns:
           Hypocenter as OpenQuake Point instance.
        """
        return point.Point(self.lon, self.lat, self.depth)

    def setMechanism(self, mech=None, rake=None, dip=None):
        """Set the earthquake mechanism manually (overriding any values read
        in from event.xml or source.txt. If rake and dip are not specified,
        they will be assigned by mechanism as follows:

        +-------+--------+-----+
        | Mech  |  Rake  | Dip |
        +=======+========+=====+
        | RS    |    90  |  40 |
        +-------+--------+-----+
        | NM    |   -90  |  50 |
        +-------+--------+-----+
        | SS    |     0  |  90 |
        +-------+--------+-----+
        | ALL   |    45  |  90 |
        +-------+--------+-----+

        Args:
            mech (str): One of 'RS' (reverse), 'NM' (normal), 'SS' (strike
                slip), or 'ALL' (unknown).
            rake (float): Value between -360 and 360 degrees. If set, will
                override default value for mechanism (see table above).
            dip (float): Value betweeen 0 and 90 degrees. If set, will override
                default value for mechanism (see table above). Value will be
                converted to range between -180 and 180 degrees.
        """
        mechs = {
            "RS": {"rake": 90.0, "dip": 40.0},
            "NM": {"rake": -90.0, "dip": 50.0},
            "SS": {"rake": 0.0, "dip": 90.0},
            "ALL": {"rake": 45.0, "dip": 90.0},
        }

        if mech not in list(mechs.keys()):
            raise ValueError(f"Mechanism must be one of: {str(list(mechs.keys()))}")

        if dip is not None:
            if dip < 0 or dip > 90:
                raise ValueError("Dip must be in range 0-90 degrees.")
        else:
            dip = mechs[mech]["dip"]

        if rake is not None:
            if rake < -180:
                rake += 360
            if rake > 180:
                rake -= 360
            if rake < -180 or rake > 180:
                raise ValueError(
                    "Rake must be transformable to be in range -180 to 180 " "degrees."
                )
        else:
            rake = mechs[mech]["rake"]

        self.dip = dip
        self.rake = rake
        self.mech = mech
        return


def write_event_file(event, xmlfile):
    """Write event.xml file.

    Args:
       event (dict): Dictionary containing fields [field, (type, example)]:

           - id (str, "us2008abcd")
           - netid (str, "us")
           - network (str, "USGS National Network")
           - lat (float, 42.1234)
           - lon (float, -85.1234)
           - depth (float, 24.1)
           - mag (float, 7.9)
           - time (HistoricTime object)
           - locstring (str, "East of the Poconos")
           - mech (str, "RS", "SS", "NM", or 'ALL') (optional)
           - reference (str, data source: 'Smith et al. 2016') (optional)
           - event_type (str, 'ACTUAL' or 'SCENARIO') (optional)
           - productcode (str, 'us2000wxyz_zoom')
           - reviewed (str, 'true' or 'false' or 'unknown') (optional)

    Returns:
        nothing: Nothing.
    """
    try:
        root = etree.Element("earthquake")
        root.attrib["id"] = event["id"]
        root.attrib["netid"] = event["netid"]
        root.attrib["network"] = event["network"]
        root.attrib["lat"] = f"{event['lat']:.4f}"
        root.attrib["lon"] = f"{event['lon']:.4f}"
        root.attrib["depth"] = f"{event['depth']:.1f}"
        root.attrib["mag"] = f"{event['mag']:.1f}"
        root.attrib["time"] = event["time"].strftime(constants.ALT_TIMEFMT)
        root.attrib["locstring"] = event["locstring"]
        if "mech" in event:
            root.attrib["mech"] = event["mech"]
        if "reference" in event:
            root.attrib["reference"] = event["reference"]
        if "productcode" in event:
            root.attrib["productcode"] = event["productcode"]
        if "event_type" in event:
            if event["event_type"] not in ["ACTUAL", "SCENARIO"]:
                raise AttributeError('event_type must be "ACTUAL" or "SCENARIO"')
            root.attrib["event_type"] = event["event_type"]
        else:
            root.attrib["event_type"] = "ACTUAL"
        if "reviewed" in event:
            root.attrib["reviewed"] = event["reviewed"]
        else:
            root.attrib["reviewed"] = "unknown"

        tree = etree.ElementTree(root)
        tree.write(xmlfile, pretty_print=True)
    except Exception as e:
        return str(e)

    return None


def read_event_file(eventxml):
    """Read event.xml file from disk, returning a dictionary of attributes.
    Input XML format looks like this (all elements are required unless
    explicitly labeled optional):

    .. code-block:: xml

         <earthquake
             id="2008ryan"
             netid="us"
             network="USGS National Network" (required but may be empty string)
             lat="30.9858"
             lon="103.3639"
             mag="7.9"
             depth="19.0"
             time="YYYY-mm-ddTHH:MM:SS.ffffffZ" (omitting fractional seconds
                                                is also supported)
             locstring="EASTERN SICHUAN, CHINA"
             mech='SS' | 'NM' | 'RS' | 'ALL' (optional)
             reference="Smith et al. 2016" (optional)
             productcode='us2008ryan' (optional)
             reviewed="true" (optional)
         />

    Args:
        eventxml (str): Path to event XML file OR file-like object.

    Returns:
       dict: Dictionary with keys:
         - id: Origin network and origin code (i.e., us2017abcd).
         - netid: Origin network ("us").
         - network: (A long-form description of the network)
         - lat: Origin latitude
         - lon: Origin longitude
         - mag: Origin magnitude
         - depth: Origin depth
         - time: Origin time as an HistoricTime object.
         - locstring: Location string
         - mech: (optional) Moment mechanism, one of:
           - 'RS' (Reverse)
           - 'SS' (Strike-Slip)
           - 'NM' (Normal)
           - 'ALL' (Undetermined)
         - reference: (optional) A description of the source of the data.
         - productcode: (optional) This product source's unique code for this
                        particular ShakeMap.
         - reviewed: Has the origin been reviewed? "true", "false", or "unknown"

    Raises:
        ValueError: If the time string cannot be parsed into a datetime object
        KeyError: If any of the required attributes are missing from event.xml
    """
    if isinstance(eventxml, str):
        tree = dET.parse(eventxml)
        root = tree.getroot()
    else:
        data = eventxml.read()
        root = dET.fromstring(data)

    # Turn XML content into dictionary
    if root.tag == "earthquake":
        xmldict = dict(root.items())
    else:
        eq = root.find("earthquake")
        xmldict = dict(eq.items())

    eqdict = {}

    #########################################################
    # A Short Primer on PDL-style Identifiers
    # Because Everybody (Including The Author) Forgets It.
    #
    # In PDL, there are 4 identifiers that fully specify a product:
    # - source The network that generated the *product* (us, ci, etc.).
    # - code   The unique ID string that identifies this product,
    #          usually prepended by *source* (us2008abcd).
    # - eventsource The network that created the *origin* (us, ci, etc.)
    # - eventsourcecode The code within that network that uniquely
    #                   identifies the event (2008abcd).
    #
    # For our purposes, we're storing *source* and *code* as
    # *productsource* and *productcode* respectively in the
    # container, in an effort to reduce confusion about their
    # meaning. Time will tell.
    #########################################################

    # read in the id fields
    eqdict["id"] = xmldict["id"]
    # This isn't optional, but maybe it isn't in some old files
    if "network" in xmldict:
        eqdict["network"] = xmldict["network"]
    else:
        eqdict["network"] = ""

    eqdict["netid"] = xmldict["netid"]

    # look for the productcode attribute in the xml,
    # otherwise use the event id
    if "productcode" in xmldict:
        eqdict["productcode"] = xmldict["productcode"]
    elif isinstance(eventxml, str):
        eqdict["productcode"] = eqdict["id"]
    else:
        # It's up to the user of this data how to construct the
        # product code
        pass

    # Support old event file date/times
    if "time" in xmldict:
        try:
            eqdict["time"] = HistoricTime.strptime(xmldict["time"], constants.TIMEFMT)
        except ValueError:
            try:
                eqdict["time"] = HistoricTime.strptime(
                    xmldict["time"], constants.ALT_TIMEFMT
                )
            except ValueError:
                raise ValueError(f"Couldn't convert {xmldict['time']} to HistoricTime")
    else:
        if (
            "year" not in xmldict
            or "month" not in xmldict
            or "day" not in xmldict
            or "hour" not in xmldict
            or "minute" not in xmldict
            or "second" not in xmldict
        ):
            raise ValueError("Missing date/time elements in event file.")
        eqdict["time"] = HistoricTime.datetime(
            xmldict["year"],
            xmldict["month"],
            xmldict["day"],
            xmldict["hour"],
            xmldict["minute"],
            xmldict["second"],
        )

    eqdict["lat"] = float(xmldict["lat"])
    eqdict["lon"] = float(xmldict["lon"])
    eqdict["depth"] = float(xmldict["depth"])
    eqdict["mag"] = float(xmldict["mag"])
    eqdict["locstring"] = xmldict["locstring"]

    if "mech" in xmldict:
        eqdict["mech"] = xmldict["mech"]
    # Older files may have "type" instead of "mech"
    if "type" in xmldict:
        eqdict["type"] = xmldict["type"]
    if "reference" in xmldict:
        eqdict["reference"] = xmldict["reference"]
    if "reviewed" in xmldict:
        eqdict["reviewed"] = xmldict["reviewed"]
    else:
        eqdict["reviewed"] = "unknown"

    return eqdict


def read_moment_quakeml(momentfile):
    """Read moment parameters from a QuakeML file.
    Args:
        momentfile (str): Path to a QuakeML file.
    Returns:
        dict: Empty if momentfile is somehow not valid, or:
              - moment:
                  - T:
                    - azimuth
                    - plunge
                  - N:
                    - azimuth
                    - plunge
                  - P:
                    - azimuth
                    - plunge
                  - NP1:
                    - strike: float
                    - dip: float
                    - rake: float
                  - NP2:
                    - strike: float
                    - dip: float
                    - rake: float
                  - type: string
                  - source: string
    """
    params = {}
    if not _is_quakeml(momentfile):
        return params

    # obspy spits out a bunch of warnings we don't care about
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        catalog = read_events(momentfile)

    if not len(catalog.events):
        return params
    event = catalog.events[0]
    if not len(event.focal_mechanisms):
        return params
    focal = catalog.events[0].focal_mechanisms[0]

    # get the source and type information, if possible
    mtype = "unknown"
    if focal.method_id is not None:
        mtype = focal.method_id

    msource = "unknown"
    if focal.creation_info is not None:
        if focal.creation_info.agency_id is not None:
            msource = focal.creation_info.agency_id

    if focal.nodal_planes is None:
        return params
    if focal.nodal_planes.nodal_plane_1 is None:
        if focal.moment_tensor.tensor.m_rr is not None:
            mrr = focal.moment_tensor.tensor.m_rr
            mtt = focal.moment_tensor.tensor.m_tt
            mpp = focal.moment_tensor.tensor.m_pp
            mrt = focal.moment_tensor.tensor.m_rt
            mrp = focal.moment_tensor.tensor.m_rp
            mtp = focal.moment_tensor.tensor.m_tp
            params = fill_tensor_from_components(
                mrr, mtt, mpp, mrt, mrp, mtp, source=msource, mtype=mtype
            )
    else:
        plane1 = focal.nodal_planes.nodal_plane_1
        plane2 = focal.nodal_planes.nodal_plane_2
        params["NP1"] = {
            "strike": plane1.strike,
            "dip": plane1.dip,
            "rake": plane1.rake,
        }
        params["NP2"] = {
            "strike": plane2.strike,
            "dip": plane2.dip,
            "rake": plane2.rake,
        }
        params["T"] = {
            "azimuth": focal.principal_axes.t_axis.azimuth,
            "plunge": focal.principal_axes.t_axis.plunge,
        }
        params["N"] = {
            "azimuth": focal.principal_axes.n_axis.azimuth,
            "plunge": focal.principal_axes.n_axis.plunge,
        }
        params["P"] = {
            "azimuth": focal.principal_axes.p_axis.azimuth,
            "plunge": focal.principal_axes.p_axis.plunge,
        }

    moment = {"moment": params}
    return moment


def read_source(sourcefile):
    """Read source.txt file, which has lines like key=value.

    Args:
        sourcefile: Path to source.txt file OR file-like object

    Returns:
        dict: Dictionary containing key/value pairs from file.
    """
    isFile = False
    if isinstance(sourcefile, str):
        f = open(sourcefile, "rt")
    else:
        isFile = True
        f = sourcefile
    params = {}
    for line in f.readlines():
        if line.startswith("#"):
            continue
        parts = line.split("=")
        key = parts[0].strip()
        value = parts[1].strip()
        # see if this is some sort of number
        try:
            value = int(value)
        except ValueError:
            try:
                value = float(value)
            except ValueError:
                pass
        params[key] = value
    if not isFile:
        f.close()
    return params



#!/usr/bin/env python

# stdlib modules
import copy

# third party imports
import numpy as np
from openquake.hazardlib.geo.point import Point
from openquake.hazardlib.geo.utils import OrthographicProjection

from esi_utils_vectors.ecef import latlon2ecef
from esi_utils_vectors.vector import Vector
from esi_utils_rupture import constants


def is_quad(q):
    """
    Checks that an individual quad is coplanar.

    Args:
        q (list): A quadrilateral; list of four OQ Points.

    Returns:
        tuple: Bool for whether or not the points are planar within tolerance;
            and also the corrected quad where p2 is adjusted to be on the same
            plane as the other points.
    """
    P0, P1, P2, P3 = q

    # Convert points to ECEF
    p0 = Vector.fromPoint(P0)
    p1 = Vector.fromPoint(P1)
    p2 = Vector.fromPoint(P2)
    p3 = Vector.fromPoint(P3)

    # Unit vector along top edge
    v0 = (p1 - p0).norm()

    # Distance along bottom edge
    d = (p3 - p2).mag()

    # New location for p2 by extending from p3 the same distance and
    # direction that p1 is from p0:
    new_p2 = p3 + v0 * d

    # How far off of the plane is the origin p2?
    planepoints = [p0, p1, p2]
    dist = get_distance_to_plane(planepoints, p2)

    # Is it close enough?
    if dist / d > constants.OFFPLANE_TOLERANCE:
        on_plane = False
    else:
        on_plane = True

    # Fixed quad
    fquad = [p0.toPoint(), p1.toPoint(), new_p2.toPoint(), p3.toPoint()]

    return (on_plane, fquad)


def get_quad_width(q):
    """
    Return width of an individual planar trapezoid, where the p0-p1 distance
    represents the long side.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        float: Width of planar trapezoid in km.
    """
    P0, P1, P2, P3 = q
    p0 = Vector.fromPoint(P0)
    p1 = Vector.fromPoint(P1)
    p3 = Vector.fromPoint(P3)
    AB = p0 - p1
    AC = p0 - p3
    t1 = (AB.cross(AC).cross(AB)).norm()
    width = t1.dot(AC) / 1000.0

    return width


def get_quad_mesh(q, dx):
    """
    Create a mesh from a quadrilateal.

    Args:
        q (list): A quadrilateral; list of four points.
        dx (float):  Target dx in km; used to get nx and ny of mesh, but mesh
            snaps to edges of rupture so actual dx/dy will not actually equal
            this value in general.

    Returns:
        dict: Mesh dictionary, which includes numpy arrays:

        - llx: lower left x coordinate in ECEF coords.
        - lly: lower left y coordinate in ECEF coords.
        - llz: lower left z coordinate in ECEF coords.
        - ulx: upper left x coordinate in ECEF coords.
        - etc.

    """
    P0, P1, P2, P3 = q
    p0 = Vector.fromPoint(P0)  # fromPoint converts to ECEF
    p1 = Vector.fromPoint(P1)
    p2 = Vector.fromPoint(P2)
    p3 = Vector.fromPoint(P3)

    # Get nx based on length of top edge, minimum allowed is 2
    toplen_km = get_quad_length(q)
    nx = int(np.max([round(toplen_km / dx, 0) + 1, 2]))

    # Get array of points along top and bottom edges
    xfac = np.linspace(0, 1, nx)
    topp = [p0 + (p1 - p0) * a for a in xfac]
    botp = [p3 + (p2 - p3) * a for a in xfac]

    # Get ny based on mean length of vectors connecting top and bottom points
    ylen_km = np.ones(nx)
    for i in range(nx):
        ylen_km[i] = (topp[i] - botp[i]).mag() / 1000
    ny = int(np.max([round(np.mean(ylen_km) / dx, 0) + 1, 2]))
    yfac = np.linspace(0, 1, ny)

    # Build mesh: dict of ny by nx arrays (x, y, z):
    mesh = {"x": np.zeros([ny, nx]), "y": np.zeros([ny, nx]), "z": np.zeros([ny, nx])}
    for i in range(nx):
        mpts = [topp[i] + (botp[i] - topp[i]) * a for a in yfac]
        mesh["x"][:, i] = [a.x for a in mpts]
        mesh["y"][:, i] = [a.y for a in mpts]
        mesh["z"][:, i] = [a.z for a in mpts]

    # Make arrays of pixel corners
    mesh["llx"] = mesh["x"][1:, 0:-1]
    mesh["lrx"] = mesh["x"][1:, 1:]
    mesh["ulx"] = mesh["x"][0:-1, 0:-1]
    mesh["urx"] = mesh["x"][0:-1, 1:]
    mesh["lly"] = mesh["y"][1:, 0:-1]
    mesh["lry"] = mesh["y"][1:, 1:]
    mesh["uly"] = mesh["y"][0:-1, 0:-1]
    mesh["ury"] = mesh["y"][0:-1, 1:]
    mesh["llz"] = mesh["z"][1:, 0:-1]
    mesh["lrz"] = mesh["z"][1:, 1:]
    mesh["ulz"] = mesh["z"][0:-1, 0:-1]
    mesh["urz"] = mesh["z"][0:-1, 1:]
    mesh["cpx"] = np.zeros_like(mesh["llx"])
    mesh["cpy"] = np.zeros_like(mesh["llx"])
    mesh["cpz"] = np.zeros_like(mesh["llx"])

    # i and j are indices over subruptures
    ni, nj = mesh["llx"].shape
    for i in range(0, ni):
        for j in range(0, nj):
            # Rupture corner points
            pp0 = Vector(mesh["ulx"][i, j], mesh["uly"][i, j], mesh["ulz"][i, j])
            pp1 = Vector(mesh["urx"][i, j], mesh["ury"][i, j], mesh["urz"][i, j])
            pp2 = Vector(mesh["lrx"][i, j], mesh["lry"][i, j], mesh["lrz"][i, j])
            pp3 = Vector(mesh["llx"][i, j], mesh["lly"][i, j], mesh["llz"][i, j])
            # Find center of quad
            mp0 = pp0 + (pp1 - pp0) * 0.5
            mp1 = pp3 + (pp2 - pp3) * 0.5
            cp = mp0 + (mp1 - mp0) * 0.5
            mesh["cpx"][i, j] = cp.x
            mesh["cpy"][i, j] = cp.y
            mesh["cpz"][i, j] = cp.z
    return mesh


def get_local_unit_slip_vector(strike, dip, rake):
    """
    Compute the components of a unit slip vector.

    Args:
        strike (float): Clockwise angle (deg) from north of the line at the
            intersection of the rupture plane and the horizontal plane.
        dip (float): Angle (deg) between rupture plane and the horizontal plane
            normal to the strike (0-90 using right hand rule).
        rake (float): Direction of motion of the hanging wall relative to the
            foot wall, as measured by the angle (deg) from the strike vector.

    Returns:
        Vector: Unit slip vector in 'local' N-S, E-W, U-D coordinates.

    """
    strike = np.radians(strike)
    dip = np.radians(dip)
    rake = np.radians(rake)
    sx = np.sin(rake) * np.cos(dip) * np.cos(strike) + np.cos(rake) * np.sin(strike)
    sy = np.sin(rake) * np.cos(dip) * np.sin(strike) + np.cos(rake) * np.cos(strike)
    sz = np.sin(rake) * np.sin(dip)
    return Vector(sx, sy, sz)


def get_local_unit_slip_vector_DS(strike, dip, rake):
    """
    Compute the DIP SLIP components of a unit slip vector.

    Args:
        strike (float): Clockwise angle (deg) from north of the line at the
            intersection of the rupture plane and the horizontal plane.
        dip (float): Angle (degrees) between rupture plane and the horizontal
            plane normal to the strike (0-90 using right hand rule).
        rake (float): Direction of motion of the hanging wall relative to the
            foot wall, as measured by the angle (deg) from the strike vector.

    Returns:
        Vector: Unit slip vector in 'local' N-S, E-W, U-D coordinates.

    """
    strike = np.radians(strike)
    dip = np.radians(dip)
    rake = np.radians(rake)
    sx = np.sin(rake) * np.cos(dip) * np.cos(strike)
    sy = np.sin(rake) * np.cos(dip) * np.sin(strike)
    sz = np.sin(rake) * np.sin(dip)
    return Vector(sx, sy, sz)


def get_local_unit_slip_vector_SS(strike, dip, rake):
    """
    Compute the STRIKE SLIP components of a unit slip vector.

    Args:
        strike (float): Clockwise angle (deg) from north of the line at the
            intersection of the rupture plane and the horizontal plane.
        dip (float): Angle (degrees) between rupture plane and the horizontal
            plane normal to the strike (0-90 using right hand rule).
        rake (float): Direction of motion of the hanging wall relative to the
            foot wall, as measured by the angle (deg) from the strike vector.

    Returns:
        Vector: Unit slip vector in 'local' N-S, E-W, U-D coordinates.

    """
    strike = np.radians(strike)
    dip = np.radians(dip)
    rake = np.radians(rake)
    sx = np.cos(rake) * np.sin(strike)
    sy = np.cos(rake) * np.cos(strike)
    sz = 0.0
    return Vector(sx, sy, sz)


def reverse_quad(q):
    """
    Reverse the verticies of a quad in the sense that the strike direction
    is flipped.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        list: Reversed quadrilateral.

    """
    return [q[1], q[0], q[3], q[2]]


def get_quad_slip(q, rake):
    """
    Compute the unit slip vector in ECEF space for a quad and rake angle.

    Args:
        q (list): A quadrilateral; list of four points.
        rake (float): Direction of motion of the hanging wall relative to
        the foot wall, as measured by the angle (deg) from the strike vector.

    Returns:
        Vector: Unit slip vector in ECEF space.

    """
    P0, P1, P2 = q[0:3]
    strike = P0.azimuth(P1)
    dip = get_quad_dip(q)
    s1_local = get_local_unit_slip_vector(strike, dip, rake)
    s0_local = Vector(0, 0, 0)
    qlats = [a.latitude for a in q]
    qlons = [a.longitude for a in q]
    proj = OrthographicProjection(
        np.min(qlons), np.max(qlons), np.min(qlats), np.max(qlats)
    )
    s1_ll = proj(np.array([s1_local.x]), np.array([s1_local.y]), reverse=True)
    s0_ll = proj(np.array([s0_local.x]), np.array([s0_local.y]), reverse=True)
    s1_ecef = Vector.fromTuple(latlon2ecef(s1_ll[1], s1_ll[0], s1_local.z))
    s0_ecef = Vector.fromTuple(latlon2ecef(s0_ll[1], s0_ll[0], s0_local.z))
    slp_ecef = (s1_ecef - s0_ecef).norm()
    return slp_ecef


def get_quad_length(q):
    """
    Length of top eduge of a quadrilateral.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        float: Length of quadrilateral in km.

    """
    P0, P1, P2, P3 = q
    p0 = Vector.fromPoint(P0)  # fromPoint converts to ECEF
    p1 = Vector.fromPoint(P1)
    qlength = (p1 - p0).mag() / 1000
    return qlength


def get_quad_dip(q):
    """
    Dip of a quadrilateral.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        float: Dip in degrees.

    """
    N = get_quad_normal(q)
    V = get_vertical_vector(q)
    dip = np.degrees(np.arccos(Vector.dot(N, V)))
    return dip


def get_quad_normal(q):
    """
    Compute the unit normal vector for a quadrilateral in
    ECEF coordinates.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        Vector: Normalized normal vector for the quadrilateral in ECEF coords.
    """
    P0, P1, P2, P3 = q
    p0 = Vector.fromPoint(P0)  # fromPoint converts to ECEF
    p1 = Vector.fromPoint(P1)
    p3 = Vector.fromPoint(P3)
    v1 = p1 - p0
    v2 = p3 - p0
    vn = Vector.cross(v2, v1).norm()
    return vn


def get_quad_strike_vector(q):
    """
    Compute the unit vector pointing in the direction of strike for a
    quadrilateral in ECEF coordinates. Top edge assumed to be horizontal.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        Vector: The unit vector pointing in strike direction in ECEF coords.
    """
    P0, P1, P2, P3 = q
    p0 = Vector.fromPoint(P0)  # fromPoint converts to ECEF
    p1 = Vector.fromPoint(P1)
    v1 = (p1 - p0).norm()
    return v1


def get_quad_down_dip_vector(q):
    """
    Compute the unit vector pointing down dip for a quadrilateral in
    ECEF coordinates.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        Vector: The unit vector pointing down dip in ECEF coords.

    """
    P0, P1, P2, P3 = q
    p0 = Vector.fromPoint(P0)  # fromPoint converts to ECEF
    p1 = Vector.fromPoint(P1)
    p0p1 = p1 - p0
    qnv = get_quad_normal(q)
    ddv = Vector.cross(p0p1, qnv).norm()
    return ddv


def get_vertical_vector(q):
    """
    Compute the vertical unit vector for a quadrilateral
    in ECEF coordinates.

    Args:
        q (list): A quadrilateral; list of four points.

    Returns:
        Vector: Normalized vertical vector for the quadrilateral in ECEF
                coords.
    """
    P0, P1, P2, P3 = q
    P0_up = copy.deepcopy(P0)
    P0_up.depth = P0_up.depth - 1.0
    p0 = Vector.fromPoint(P0)  # fromPoint converts to ECEF
    p1 = Vector.fromPoint(P0_up)
    v1 = (p1 - p0).norm()
    return v1


def _quad_distance(q, points, horizontal=False):
    """
    Calculate the shortest distance from a set of points to a rupture surface.

    Args:
        q (list): A quadrilateral; list of four points.
        points (array): Numpy array Nx3 of points (ECEF) to calculate distance
            from.
        horizontal:  Boolean indicating whether to treat points inside quad as
            0 distance.

    Returns:
        float: Array of size N of distances (in km) from input points to
            rupture surface.
    """
    P0, P1, P2, P3 = q

    if horizontal:
        # project this quad to the surface
        P0 = Point(P0.x, P0.y, 0)
        P1 = Point(P1.x, P1.y, 0)
        P2 = Point(P2.x, P2.y, 0)
        P3 = Point(P3.x, P3.y, 0)

    # Convert to ecef
    p0 = Vector.fromPoint(P0)
    p1 = Vector.fromPoint(P1)
    p2 = Vector.fromPoint(P2)
    p3 = Vector.fromPoint(P3)

    # Make a unit vector normal to the plane
    normalVector = (p1 - p0).cross(p2 - p0).norm()

    dist = np.ones_like(points[:, 0]) * np.nan

    p0d = p0.getArray() - points
    p1d = p1.getArray() - points
    p2d = p2.getArray() - points
    p3d = p3.getArray() - points

    # Create 4 planes with normals pointing outside rectangle
    n0 = (p1 - p0).cross(normalVector).getArray()
    n1 = (p2 - p1).cross(normalVector).getArray()
    n2 = (p3 - p2).cross(normalVector).getArray()
    n3 = (p0 - p3).cross(normalVector).getArray()

    sgn0 = np.signbit(np.sum(n0 * p0d, axis=1))
    sgn1 = np.signbit(np.sum(n1 * p1d, axis=1))
    sgn2 = np.signbit(np.sum(n2 * p2d, axis=1))
    sgn3 = np.signbit(np.sum(n3 * p3d, axis=1))

    inside_idx = (sgn0 == sgn1) & (sgn1 == sgn2) & (sgn2 == sgn3)
    if horizontal:
        dist[inside_idx] = 0.0
    else:
        dist[inside_idx] = np.power(
            np.abs(np.sum(p0d[inside_idx, :] * normalVector.getArray(), axis=1)), 2
        )

    outside_idx = np.logical_not(inside_idx)
    s0 = _distance_sq_to_segment(p0d, p1d)
    s1 = _distance_sq_to_segment(p1d, p2d)
    s2 = _distance_sq_to_segment(p2d, p3d)
    s3 = _distance_sq_to_segment(p3d, p0d)

    smin = np.minimum(np.minimum(s0, s1), np.minimum(s2, s3))
    dist[outside_idx] = smin[outside_idx]
    dist = np.sqrt(dist) / 1000.0
    shp = dist.shape
    if len(shp) == 1:
        dist.shape = (shp[0], 1)
    if np.any(np.isnan(dist)):
        raise Exception("Could not calculate some distances!")
    dist = np.fliplr(dist)
    return dist


def get_distance_to_plane(planepoints, otherpoint):
    """
    Calculate a point's distance to a plane.  Used to figure out if a
    quadrilateral points are all co-planar.

    Args:
        planepoints (list): List of three points (from Vector class) defining a
            plane.
        otherpoint (Vector): 4th Vector to compare to points defining the
            plane.

    Returns:
        float: Distance (in meters) from otherpoint to plane.

    """
    # from
    # https://en.wikipedia.org/wiki/Plane_(geometry)#Describing_a_plane_through_three_points
    p0, p1, p2 = planepoints
    x1, y1, z1 = p0.getArray()
    x2, y2, z2 = p1.getArray()
    x3, y3, z3 = p2.getArray()
    D = np.linalg.det(np.array([[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]]))
    if D != 0:
        d = -1
        at = np.linalg.det(np.array([[1, y1, z1], [1, y2, z2], [1, y3, z3]]))
        bt = np.linalg.det(np.array([[x1, 1, z1], [x2, 1, z2], [x3, 1, z3]]))
        ct = np.linalg.det(np.array([[x1, y1, 1], [x2, y2, 1], [x3, y3, 1]]))
        a = (-d / D) * at
        b = (-d / D) * bt
        c = (-d / D) * ct

        numer = np.abs(a * otherpoint.x + b * otherpoint.y + c * otherpoint.z + d)
        denom = np.sqrt(a**2 + b**2 + c**2)
        dist = numer / denom
    else:
        dist = 0
    return dist


def _distance_sq_to_segment(p0, p1):
    """
    Calculate the distance^2 from the origin to a segment defined by two
    vectors

    Args:
        p0 (array): Numpy array Nx3 (ECEF).
        p1 (array): Numpy array Nx3 (ECEF).

    Returns:
        float: The squared distance from the origin to a segment.
    """
    # /*
    #  * This algorithm is from (Vince's) CS1 class.
    #  * It returns the distance^2 from the origin to a segment defined
    #  * by two vectors
    #  */

    dist = np.zeros_like(p1[:, 0])
    # /* Are the two points equal? */
    idx_equal = (p0[:, 0] == p1[:, 0]) & (p0[:, 1] == p1[:, 1]) & (p0[:, 2] == p1[:, 2])
    dist[idx_equal] = np.sqrt(
        p0[idx_equal, 0] ** 2 + p0[idx_equal, 1] ** 2 + p0[idx_equal, 2] ** 2
    )

    v = p1 - p0

    # /*
    #  * C1 = c1/|v| is the projection of the origin O on line (P0,P1).
    #  * If C1 is negative, then O is outside the segment and
    #  * closer to the P0 side.
    #  * If C1 is positive and >V then O is on the other side.
    #  * If C1 is positive and <V then O is inside.
    #  */

    c1 = -1 * np.sum(p0 * v, axis=1)
    idx_neg = c1 <= 0
    dist[idx_neg] = p0[idx_neg, 0] ** 2 + p0[idx_neg, 1] ** 2 + p0[idx_neg, 2] ** 2

    c2 = np.sum(v * v, axis=1)
    idx_less_c1 = c2 <= c1
    dist[idx_less_c1] = (
        p1[idx_less_c1, 0] ** 2 + p1[idx_less_c1, 1] ** 2 + p1[idx_less_c1, 2] ** 2
    )

    idx_other = np.logical_not(idx_neg | idx_equal | idx_less_c1)

    nr, nc = p0.shape
    t1 = c1 / c2
    t1.shape = (nr, 1)
    tmp = p0 + (v * t1)
    dist[idx_other] = (
        tmp[idx_other, 0] ** 2 + tmp[idx_other, 1] ** 2 + tmp[idx_other, 2] ** 2
    )

    return dist


def rake_to_mech(rake):
    """
    Convert rake to mechanism.

    Args:
        rake (float): Rake angle in degrees.

    Returns:
        str: Mechanism.
    """
    mech = "ALL"
    if rake is not None:
        if (
            (rake >= -180 and rake <= -150)
            or (rake >= -30 and rake <= 30)
            or (rake >= 150 and rake <= 180)
        ):
            mech = "SS"
        if rake >= -120 and rake <= -60:
            mech = "NM"
        if rake >= 60 and rake <= 120:
            mech = "RS"
    return mech



#!/usr/bin/env python

# stdlib imports
import copy

# third party imports
import numpy as np

from openquake.hazardlib.gsim.base import GMPE
from esi_shakelib.utils.contexts import DistancesContext

# local imports
from esi_utils_rupture.edge_rupture import EdgeRupture


class Distance(object):
    """
    Class for distance calculations. Primary method is 'get_distance'. To
    handle multiple segment ruptures, many of the distances are based on the
    Spudich and Chiou (2015) GC2 coordinate system.


    References:
        Spudich, Paul and Chiou, Brian, 2015, Strike-parallel and strike-normal
        coordinate system around geometrically complicated rupture traces—Use
        by NGA-West2 and further improvements: U.S. Geological Survey Open-File
        Report 2015-1028, 20 p., http://dx.doi.org/10.3133/ofr20151028.
    """

    def __init__(self, gmpe, lon, lat, dep, rupture=None):
        """
        Constructor for Distance class.

        Args:
            gmpe (GMPE): Concrete subclass of GMPE
                `[link] <http://docs.openquake.org/oq-hazardlib/master/gsim/index.html#built-in-gsims>`__;
                can be individual instance or list of instances.
            lon (array): A numpy array of site longitudes.
            lat (array): A numpy array of site latitudes.
            dep (array): A numpy array of site depths (km); down is positive.
            rupture (Rupture): A Shakemap Rupture instance.

        Returns:
            Distance object.
        """  # noqa

        self._rupture = rupture

        self._distance_context = self._calcDistanceContext(gmpe, lat, lon, dep)
        self._distance_context.lons = lon
        self._distance_context.lats = lat

    @classmethod
    def fromSites(cls, gmpe, sites, rup):
        """
        Convenience class method to construct a Distance object from a sites
        object.

        Args:
            gmpe (GMPE): Concrete subclass of GMPE
                `[link] <http://docs.openquake.org/oq-hazardlib/master/gsim/index.html#built-in-gsims>`__;
                can be individual instance or list of instances.
            sites (Sites): A Shakemap Sites object.
            rup (Rupture): A Shakemap Rupture object.

        Returns:
            Distance object.
        """  # noqa
        lon, lat = np.meshgrid(sites._lons, sites._lats)
        dep = np.zeros_like(lon)
        return cls(gmpe, lon, lat, dep, rup)

    def getDistanceContext(self):
        """
        Returns:
            Openquake distance context
            `[link] <http://docs.openquake.org/oq-hazardlib/master/gsim/index.html?highlight=distancescontext#openquake.hazardlib.gsim.base.DistancesContext>`__.
        """  # noqa
        return copy.deepcopy(self._distance_context)

    def _calcDistanceContext(self, gmpe, lat, lon, dep):
        """
        Create a DistancesContext object.

        Args:
            gmpe (GMPE): Concrete subclass of GMPE
                (https://github.com/gem/oq-hazardlib/blob/master/openquake/hazardlib/gsim/base.py)
                can be individual instance or list of instances.
            lat (array): Numpy array of latitudes.
            lon (array): Numpy array of longitudes.
            dep (array): Numpy array of depths (km).

        Returns:
            DistancesContext object with distance grids required by input
            gmpe(s).

        Raises:
            TypeError: If gmpe is not a subclass of GMPE.
        """
        if not isinstance(gmpe, list):
            gmpe = [gmpe]

        # require rhypo always
        requires = set(["repi", "rhypo", "rrup", "rjb"])

        for ig in gmpe:
            if not isinstance(ig, GMPE):
                raise TypeError(
                    "getDistanceContext() cannot work with objects of "
                    f'type "{type(ig)}"'
                )
            requires = requires | ig.REQUIRES_DISTANCES

        context = DistancesContext()

        if isinstance(self._rupture, EdgeRupture):
            ddict = get_distance(
                list(requires), lat, lon, dep, self._rupture, dx=self._rupture._mesh_dx
            )
        else:
            ddict = get_distance(list(requires), lat, lon, dep, self._rupture)
        for method in requires:
            (context.__dict__)[method] = ddict[method]

        return context


def get_distance_measures():
    """
    Returns a list of strings specifying the distance measure types
    (e.g. "repi", "rhypo") available for the "methods" argument to
    get_diistance().

    Returns:
        A list of strings.
    """

    return ["repi", "rhypo", "rjb", "rrup", "rx", "ry", "ry0", "U", "T", "rvolc"]


def get_distance(methods, lat, lon, dep, rupture, dx=0.5):
    """
    Calculate distance using any one of a number of distance measures.
    One of quadlist OR hypo must be specified. The following table gives
    the allowed distance strings and a description of each.

    +--------+----------------------------------------------------------+
    | String | Description                                              |
    +========+==========================================================+
    | repi   | Distance to epicenter.                                   |
    +--------+----------------------------------------------------------+
    | rhypo  | Distance to hypocenter.                                  |
    +--------+----------------------------------------------------------+
    | rjb    | Joyner-Boore distance; this is closest distance to the   |
    |        | surface projection of the rupture plane.                 |
    +--------+----------------------------------------------------------+
    | rrup   | Rupture distance; closest distance to the rupture plane. |
    +--------+----------------------------------------------------------+
    | rx     | Strike-normal distance; same as GC2 coordiante T.        |
    +--------+----------------------------------------------------------+
    | ry     | Strike-parallel distance; same as GC2 coordiante U, but  |
    |        | with a shift in origin definition. See Spudich and Chiou |
    |        | (2015) http://dx.doi.org/10.3133/ofr20151028.            |
    +--------+----------------------------------------------------------+
    | ry0    | Horizontal distance off the end of the rupture measured  |
    |        | parallel to strike. Can only be zero or positive. We     |
    |        | compute this as a function of GC2 coordinate U.          |
    +--------+----------------------------------------------------------+
    | U      | GC2 coordinate U.                                        |
    +--------+----------------------------------------------------------+
    | T      | GC2 coordinate T.                                        |
    +--------+----------------------------------------------------------+
    | rvolc  | Part of the rupture distance spent in volcanic region.   |
    |        | Currently set to 0 for all points.                       |
    +--------+----------------------------------------------------------+

    Args:
        methods (list): List of strings (or just a string) of distances to
            compute.
        lat (array): A numpy array of latitudes.
        lon (array): A numpy array of longidues.
        dep (array): A numpy array of depths (km).
        rupture (Rupture): A ShakeMap Rupture instance.
        dx (float): Mesh spacing for rupture; only used if rupture is an
            EdgeRupture subclass.

    Returns:
       dict: dictionary of numpy arrays of distances, size of lon.shape.
    """
    rupture._mesh_dx = dx

    # Dictionary for holding/returning the requested distances
    distdict = dict()

    # Coerce methods into list if it isn't
    if not isinstance(methods, list):
        methods = [methods]

    # Check that all requested distances are available
    methods_available = set(get_distance_measures())
    if not set(methods).issubset(methods_available):
        raise NotImplementedError(
            "One or more requested distance method is not "
            "valid or is not implemented yet"
        )

    # Check dimensions of site coordinates
    if (lat.shape != lon.shape) or (lat.shape != dep.shape):
        raise ValueError("lat, lon, and dep must have the same shape.")

    # -------------------------------------------------------------------------
    # Point distances
    # -------------------------------------------------------------------------
    if "rhypo" in methods:
        distdict["rhypo"] = rupture.computeRhyp(lon, lat, dep)

    if "repi" in methods:
        distdict["repi"] = rupture.computeRepi(lon, lat, dep)

    # -------------------------------------------------------------------------
    # Rupture distances
    # -------------------------------------------------------------------------
    gc2_distances = set(["rx", "ry", "ry0", "U", "T"])
    if "rrup" in methods:
        rrup = rupture.computeRrup(lon, lat, dep)
        distdict["rrup"] = rrup

    if "rjb" in methods:
        rjb = rupture.computeRjb(lon, lat, dep)
        distdict["rjb"] = rjb

    # If any of the GC2-related distances are requested, may as well do all
    if len(set(methods).intersection(gc2_distances)) > 0:
        distdict.update(rupture.computeGC2(lon, lat, dep))

    if "rvolc" in methods:
        distdict["rvolc"] = np.zeros_like(lon)

    return distdict
